#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""M-tortuosity computation on 3D binary image calling IPSDK library for the seededDistanceMap3dImg.
   Author : francois.cokelaer@ifpen.fr
"""

import os
import math
import argparse
import numpy as np
import time
import functools
from typing import Dict, Any
import logging
from mtortuosity.utils import euclidean_dist, center_of_mass, get_seed_points, logger
from mtortuosity.plugim_io import read_fda, write_fda, get_all_params_from_xml

try:
    import PyIPSDK
    import PyIPSDK.IPSDKIPLAdvancedMorphology as advmorpho
    import PyIPSDK.IPSDKIPLUtility as util
    import PyIPSDK.IPSDKIPLBinarization as bin
    import PyIPSDK.IPSDKIPLArithmetic as arithm
    import PyIPSDK.IPSDKIPLLogical as logic
except ImportError as err:
    logger.error("Your IPSDK license has expired. Please contact IPSDK support.")
    raise err


def dump_config(config: Dict[str, Any]):
    """ "
    Dump input configuration
    """
    logger.info(f"------------ m-tortuosity : Configuration----")
    logger.info(f"- root_path : {config['root_path']} ----")
    logger.info(f"- input_image : {config['input_image']} ----")
    logger.info(f"- n_points : {config['n_points']} ----")
    logger.info(f"- random seed : {config['seed']} ----")
    logger.info(f"- mode : {config['mode']} ----")
    logger.info(
        f"- csv_points_file : {config.get('csv_points_file', 'no csv_points_file provided')} ----"
    )
    if config.get("plugim", False):
        logger.info(f"------------ PLUGIM configuration ----")
        logger.info(f"- IN : {config['IN']} ----")
        logger.info(f"- OUT : {config['OUT']} ----")
        logger.info(f"- OUTPREV : {config['OUTPREV']} ----")
        logger.info(f"- result_directory : {config['result_directory']} ----")
    logger.info(f"------------------------------")


def logger_add_filehandler(filename):
    file_handler = logging.FileHandler(filename, mode="w")
    file_handler.setLevel(logging.INFO)
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    file_handler.setFormatter(formatter)
    logger.addHandler(file_handler)


def m_tortuosity_binary(config: Dict[str, Any]) -> float:
    """ "
    Compute m_tortuosity on a binary image
    """

    start = time.time()

    logger_add_filehandler(
        os.path.join(config["result_directory"], "M-tortuosity_results.txt")
    )

    dump_config(config)

    # For reproductibility
    np.random.seed(config["seed"])

    # Load input (binary) image
    if config.get("plugim", False):
        in_binary_array = read_fda(config["IN"])
        in_binary = PyIPSDK.fromArray(in_binary_array)
    else:
        in_binary = PyIPSDK.loadTiffImageFile(
            os.path.join(config["root_path"], config["input_image"])
        )
    
    # Convert to appropriate IPSDK data type
    in_binary = bin.lightThresholdImg(in_binary, 1.0)

    # Random points sampling
    points = get_seed_points(config, in_binary.array.shape)

    # Allocate working array
    working_array = np.zeros(in_binary.array.shape, dtype=bool)
    seed_binary = PyIPSDK.fromBinaryArray(working_array)

    # Find valid points set
    valid_points = set(filter(lambda p: in_binary.array[p] != 0, points))
    if len(points - valid_points):
        logger.info(
            f"Points {points-valid_points}  discarded because resides in background."
        )
    logger.info(f"Valid seed Points list for computation :  {valid_points}")
    """if len(valid_points) == 1:
        logger.error(
            f"Only one valid point after filtering, please change your inputs."
        )
        raise Exception(
            f"Only one valid point after filtering, please change your inputs."
        )"""

    # Weight for IPSK genericSeededDistanceMap3dImg ( scale x10)
    scale_factor = 10
    inDistWeight3d = PyIPSDK.createDistWeight3d(
        scale_factor,
        scale_factor,
        scale_factor,
        scale_factor * math.sqrt(2),
        scale_factor * math.sqrt(2),
        scale_factor * math.sqrt(2),
        scale_factor * math.sqrt(3),
    )

       

    c_n = []
    for p in valid_points:
        logger.info(f"M-coefficient computation for point {p}")
        # Init seed
        seed_binary.array[p] = 1
        # Seeded distance map computation
        outDistImg = advmorpho.genericSeededDistanceMap3dImg(
            in_binary, seed_binary, inDistWeight3d
        )
        if config.get("plugim", False) and config["procExecute"]:
            write_fda(outDistImg.array.astype(np.uint32), config["OUTPREV"])

        # downscale distances map result by 10
        #outDistImg.array[:] = outDistImg.array[:] / scale_factor
        outDistImg = arithm.multiplyScalarImg(outDistImg, 1/scale_factor)

        # Tortuosity computation from p for all pp
        pp_tortuosity = [
            (
                pp,
                outDistImg.array[pp] / euclidean_dist(p, pp),
            )
            for pp in valid_points
            if (pp != p and outDistImg.array[pp] != 0)
                
        ]

        # # M-coefficient computation for point p :
        num = sum([(1.0 / outDistImg.array[pp]) for pp, _ in pp_tortuosity])
        denom = sum(
            [
                1.0 / (outDistImg.array[pp] * tortuosity)
                for pp, tortuosity in pp_tortuosity
            ]
        )
        m_coeff = num / denom if denom != 0 else 0
        c_n.append(m_coeff)
        logger.info(f"M-coefficient for point {p} result is {m_coeff}")

        # re-init seed image for next seed point
        seed_binary.array[p] = 0

    # M-scalar computation
    logger.info(f"M-scalar computation:")
    center = center_of_mass(in_binary)
    logger.info(f"Center of Mass is {center}")
    num = sum(
        [
            1.0 / euclidean_dist(p, center)
            for idx_p, p in enumerate(valid_points)
            if c_n[idx_p] != 0
        ]
    )
    denom = sum(
        [
            1.0 / (euclidean_dist(p, center) * c_n[idx_p])
            for idx_p, p in enumerate(valid_points)
            if c_n[idx_p] != 0
        ],
    )
    m_scalar = num / denom if denom != 0 else 0
    logger.info(f"M-scalar result: {m_scalar}")
    end = time.time()
    logger.info(f"Elapsed time is : {(end - start)}")
    return m_scalar


def main():
    parser = argparse.ArgumentParser("M Tortuosity 3D : BINARY mode only")
    parser.add_argument(
        "--plugim",
        action="store_true",
        help="Use this flag for PLUGIM with a *PAR.xml file",
    )
    parser.add_argument(
        "--root_path",
        type=str,
        default="C:\\Users\\ae\\AppData\\Local\\IFPEN\\plugim\\PlugIns\\IPSDKMTortuosity\\_Help",
        help="Root path where image and see points files can be found.",
    )
    parser.add_argument(
        "--input_image", type=str, default="3D.tif", help="Image filename."
    )
    parser.add_argument(
        "--mode", choices=["random", "deterministic"], default="deterministic"
    )
    parser.add_argument(
        "--n_points",
        type=int,
        default=5,
        help="For random mode, number of randomly selected seed points from where the m-tortuosity is computed.",
    )
    parser.add_argument(
        "--seed", type=int, default=32, help="Numpy random seed for reproductibility."
    )

    parser.add_argument(
        "--csv_points_file",
        type=str,
        default="det_points.csv",
        help="Input seed points coordinates for deterministic mode.",
    )

    args = parser.parse_args()
    if args.plugim:
        logger.info(f"Running with Plugim Mode ENABLED")
        config = get_all_params_from_xml(
            os.path.join(os.getcwd(), "IPSDKMTortuosityPAR.xml")
        )
        config["mode"] = "random"
        config["input_image"] = config["IN"]
        config["root_path"] = os.path.dirname(config["IN"])
        config["n_points"] = int(config["n_points"])
        config["seed"] = int(config["seed"])
        config["plugim"] = True
        m_tortuosity_binary(config)
    else:
        logger.info(f"Running with Plugim Mode DISABLED")
        if args.mode == "deterministic":
            assert args.csv_points_file is not None
            if args.n_points:
                logger.warning("n_points argument is discarded in deterministic mode")
        args.result_directory = "."
        m_tortuosity_binary(vars(args))


if __name__ == "__main__":
    main()
