import os
import sys
import functools
import logging
import math
from typing import Tuple, List, Dict, Any, Set
import numpy as np
import pandas as pd
import PyIPSDK.IPSDKIPLGlobalMeasure as glbmsr
import PyIPSDK.IPSDKIPLArithmetic as arithm


def setup_logger() -> Any:
    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)

    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    handler.setFormatter(formatter)

    logger.addHandler(handler)
    return logger


logger = setup_logger()


def sampling(shape: Tuple[int], n_points: int) -> Set[Tuple[Any, ...]]:
    points = set(
        tuple(map(lambda size: int(np.random.uniform(0, size)), shape))
        for _ in range(n_points)
    )
    return points


def euclidean_dist(p1: Tuple[Any, ...], p2: Tuple[Any, ...]) -> float:
    return math.sqrt(sum(map(lambda elt: pow(elt[0] - elt[1], 2), zip(p1, p2))))


def center_of_mass(in_binary : Any) -> Tuple[int]:
    outFormulaX = arithm.formula3dImg("x*I1", InOptImg3d1=in_binary)
    outFormulaY = arithm.formula3dImg("y*I1", InOptImg3d1=in_binary)
    outFormulaZ = arithm.formula3dImg("z*I1", InOptImg3d1=in_binary)
    statFormulaX = glbmsr.statsMsr3d(outFormulaX)
    statFormulaY = glbmsr.statsMsr3d(outFormulaY)
    statFormulaZ = glbmsr.statsMsr3d(outFormulaZ)
    statBinary = glbmsr.statsMsr3d(in_binary)
    x=statFormulaX.sum / statBinary.sum
    y=statFormulaY.sum / statBinary.sum
    z=statFormulaZ.sum / statBinary.sum
    return x,y,z



def parse_csv_points_file(csv_points_file: str) -> Set[Tuple[Any, ...]]:
    points_df = pd.read_csv(csv_points_file)
    assert "z" in points_df.columns
    assert "y" in points_df.columns
    assert "x" in points_df.columns
    return set(map(lambda point: tuple(point), points_df.values.tolist()))


def get_seed_points(config: Dict[str, Any], shape: Tuple[int]) -> Set[Tuple[Any, ...]]:
    if config["mode"] == "deterministic":
        deterministic_points = parse_csv_points_file(
            os.path.join(config["root_path"], config["csv_points_file"])
        )

        def _check_coordinates(point):
            return all(p < s and p >= 0 for p, s in zip(point, shape))

        filtered_set = set(
            filter(lambda p: _check_coordinates(p), deterministic_points)
        )
        if len(deterministic_points - filtered_set):
            logger.info(
                f"Points {(deterministic_points - filtered_set)}  discarded because resides out of image bounds."
                f"Image shape is {shape}"
            )
        return filtered_set
    elif config["mode"] == "random":
        return sampling(shape, config["n_points"])
    else:
        raise ValueError("Mode should be either deterministic or random")
